import logging

LEVEL = logging.DEBUG
VERBOSE = (LEVEL == logging.DEBUG)
CORE_LOGGER = logging.getLogger('sptp')
FORMAT = '%(relativeCreated)-10d[ms] %(name)-30s:%(levelname)-5s: %(message)s'

STREAM_HANDLER = logging.StreamHandler()
STREAM_HANDLER.setFormatter(logging.Formatter(FORMAT))



def _logger(name):
    return CORE_LOGGER.getChild(name)

def setup_logging():
    logging.basicConfig(format=FORMAT)

    CORE_LOGGER.setLevel(LEVEL)
    # if STREAM_HANDLER not in CORE_LOGGER.handlers:
    #     CORE_LOGGER.addHandler(STREAM_HANDLER)


import os
from dataclasses import dataclass


class SpintopEnv(object):
    ENV_NAMES = [
        'SPINTOP_MONGO_URI',
        'SPINTOP_DATABASE_NAME',
        'SPINTOP_POSTGRES_URI',
        'SPINTOP_API_URL'
    ]

    def __init__(self, init_values=None):
        if init_values is None:
            init_values = {}
        self._values = init_values

    def __getitem__(self, key):
        if key not in self.ENV_NAMES:
            raise KeyError(key)
        
        try:
            return self._values[key]
        except KeyError:
            return os.environ[key]
    
    def __setitem__(self, key, value):
        self._values[key] = value

    def freeze(self):
        return {key: self[key] for key in self.ENV_NAMES}

    



from spintop.generators import Generator

from spintop.models import (
    SpintopSerializedFlatTestRecord,
    TestRecordSummary, 
    FeatureRecord, 
    SpintopTestRecord,
    Query
)


from ..logs import _logger
from ..errors import SpintopException

logger = _logger('persistence')


class MissingMapper(SpintopException):
    def __init__(self, cls):
        super().__init__("Mapper for class {!r} is mandatory.".format(cls))

class NoMapperForObject(SpintopException):
    def __init__(self, obj, mappers):
        super(NoMapperForObject, self).__init__(
            'There are no known mapper able to interact with obj {!r} of class {}. Declared mappers are: {}'.format(
                obj,
                obj.__class__,
                [cls.__name__ for cls in mappers]
            )
        )
        
class DuplicateMapperClassName(SpintopException):
    def __init__(self, objcls):
        super(DuplicateMapperClassName, self).__init__(
            'The name of the class {} is duplicate. The class name linked to a mapper must be unique.'.format(
                objcls,
            )
        )

class TestRecordNotFound(SpintopException):
    def __init__(self, test_uuid):
        super().__init__(f'Test Record with UUID {test_uuid!r} not found.')

class ManyTestRecordFound(SpintopException):
    def __init__(self, test_uuid, count):
        super().__init__(f'{count} Test Records where found with UUID {test_uuid!r}, but only one was expected.')


class PersistenceFacade(object):
    logger = logger
    def __init__(self, serializer):
        self.serializer = serializer

    @classmethod
    def from_env(cls, env=None):
        raise NotImplementedError()

    def serialize_barrier(self, records):
        records = list(records)
        if records and isinstance(records[0], SpintopTestRecord):
            records = [self.serializer.serialize(obj) for obj in records]

        if records and isinstance(records[0], dict):
            return [SpintopSerializedFlatTestRecord(**record) for record in records]
        else:
            return records

    def create(self, records):
        records = self.serialize_barrier(records)
        return self._create(records)

    def _create(self, records):
        raise NotImplementedError()
        
    def retrieve(self, query=None, deserialize=True, limit_range=None):
        serialized_collection = self._retrieve(query, limit_range=limit_range)
        if deserialize:
            yield from serialized_collection.deserialize(self.serializer)
        else:
            yield from (record for record in serialized_collection.records)

    def _retrieve(self, query, limit_range=None):
        """Should return a ``SpintopSerializedTestRecordCollection``"""
        raise NotImplementedError()
        
    def retrieve_one(self, test_uuid, deserialize=True):
        records = self.retrieve(Query().test_uuid_is(test_uuid), deserialize=deserialize)
        records = list(records)

        count = len(records)
        if count < 1:
            raise TestRecordNotFound(test_uuid)
        elif count > 1:
            raise ManyTestRecordFound(test_uuid, count)
        else:
            return records[0]
    
    def update(self, records, upsert=False):
        records = self.serialize_barrier(records)
        return self._update(records, upsert=upsert)

    def _update(self, records, upsert=False):
        raise NotImplementedError()
    
    def delete(self, query):
        return self._delete(query)

    def _delete(self, match_query):
        raise NotImplementedError()
    
    def create_records_generator(self):
        return PersistenceGenerator(self)
        
def create_mapper_name_index(mappers):
    mappers_name_index = {}
    for mapped_cls, mapper in mappers.items():
        name = mapped_cls.__name__
        if name in mappers_name_index:
            raise DuplicateMapperClassName(mapped_cls)
        mappers_name_index[name] = mapper
    return mappers_name_index
    
class PersistenceGenerator(Generator):
    def __init__(self, facade):
        super().__init__()
        self.facade = facade
    
    def __call__(self, *args, **kwargs):
        return self.facade.retrieve(*args, **kwargs)

class Mapper(object):
    def init(self):
        pass


from sqlalchemy import or_, and_

from spintop.models import TestRecordSummary

class PostgreSQLQuery(object):
    def __init__(self, query):
        self.orig_query = query
    
    def build_data_query(self, test_record_table):
        """Only support test_uuids for now"""
        test_uuids = []

        for key, value in self.orig_query.value_equals.items():
            if key == TestRecordSummary.test_id.test_uuid.name_:
                test_uuids.append(value)
            else:
                raise NotImplementedError('PostgreSQL QuerySet only supports querying of the test_uuid field.')
        
        for key, value in self.orig_query.list_contains.items():
            raise NotImplementedError()

        for key, list_of_values in self.orig_query.value_equals_one_of.items():
            if key == TestRecordSummary.test_id.test_uuid.name_:
                test_uuids += list_of_values
            else:
                raise NotImplementedError('PostgreSQL QuerySet only supports querying of the test_uuid field.')
        
        if test_uuids:
            return [test_record_table.test_uuid.in_(test_uuids)]
        else:
            return []


from collections import Mapping
from contextlib import contextmanager
import cProfile, pstats, io

import yaml

def load_yaml_file(filename):
    with open(filename,'r') as readfile:
        data = readfile.read()
        return yaml.safe_load(data)
    

@contextmanager
def profiling(output_filename):
    pr = cProfile.Profile()
    pr.enable()
    try:
        yield
    finally:
        pr.disable()
        s = io.StringIO()
        sortby = 'cumulative'
        with open(output_filename, 'w+') as profile_stream:
            ps = pstats.Stats(pr, stream=profile_stream).sort_stats(sortby)
            ps.print_stats()

class GetRecursiveMixin(object):
    
    def get_recursive(self, as_str=None, as_array=None):
        if as_str:
            as_array = as_str.split('.')
        
        this_attr_name, *as_array = as_array
        this_attr = self._get_attr_recursive(this_attr_name)
        

        if as_array:
            if not isinstance(this_attr, GetRecursiveMixin):
                this_attr = AnonymousGetRecursive(this_attr)
            return this_attr.get_recursive(as_array=as_array)
        else:
            return this_attr
    
    def _get_attr_recursive(self, name):
        return getattr(self, name)

class GetRecursiveKeyMixin(GetRecursiveMixin):

    def _get_attr_recursive(self, name):
        return self[name]

class AnonymousGetRecursive(GetRecursiveMixin):
    def __init__(self, obj):
        self.obj = obj

    def _get_attr_recursive(self, name):
        if isinstance(self.obj, Mapping):
            return self.obj[name]
        else:
            return getattr(self.obj, name)
